#include "opencv2/opencv.hpp"
#include <iomanip>
#include <unistd.h>

using namespace std;
using namespace cv;
/* Triangulate
 * Given the list ofthe camera matrices and the list of the 2D images projections of the cameras, outputs a triangulated 3D point.
 */
int Triangulate(const vector<Mat> PM, const vector<int> index, const vector<Point2f> projPoints, Mat& X)
{
    /* parameters
   - PM : list of all 3x4 camera matrices
   - index : list of cam numbers corresponding to projections
   - projPoints : list of 2D image projections
   - points3D : triangulated 3D point (output)
    */

    // Check if enough views to triangulate
    int nviews=index.size();
    if (nviews<2) return(0);

    Mat A, P, r;
    float x,y;

    // Fill matrix A
    for( int j = 0; j < nviews; j++ ) {

        // get image projection
        x = (float) projPoints[j].x;
        y = (float) projPoints[j].y;

        // get corresponding cam matrix
        P = PM[index[j]];

        // add 3 lines to A (cf HZ book chap 12.2 p312)reference camera with more tha 2 views 3d
        r = P.row(2)*x - P.row(0); A.push_back(r);
        r = P.row(2)*y - P.row(1); A.push_back(r);
        // r = x*P.row(1) - y*P.row(0);  A.push_back(r); // redundant
    }

    // Solve homogenous system AX=0 with |X|=1, using Singular Value Decomposition
    // ref  Hartley&Zissermann book p592
    SVD::solveZ(A, X);

    // Perform reweighting and iterate solving to increase accuracy
    // ref Hartley&Sturm 1997 paper "Triangulation", Iterative-Eigen method
    float w;
    for (int c=0; c<5; c++) { // iterate 5 times
        for( int j = 0; j < nviews; j++ ) {
            int i=2*j;
            r = PM[index[j]].row(2)*X;
            w=r.at<float>(0);
            if (abs(w)<1e-6) break;
            w=1/w;
            r=A.row(i); A.row(i)=r*w;
            r=A.row(i+1); A.row(i+1)=r*w;
        }
        SVD::solveZ(A, X);
    }
    //cout<<X<<endl<<endl<<endl;
    if (!checkRange(X) || X.at<float>(3)==0) {
        cout << "SVD algorithm error" << endl;
        return(0);
    }
    else {

        // return real (inhomogeneous) 3D point
        X.at<float>(0) = X.at<float>(0) / X.at<float>(3);
        X.at<float>(1) = X.at<float>(1) / X.at<float>(3);
        X.at<float>(2) = X.at<float>(2) / X.at<float>(3);
        X.at<float>(3) = X.at<float>(3) / X.at<float>(3);
        //  cout<<"here"<<endl;
    }

    return(1);
}


